"""
    sphinxnotes.any.domain
    ~~~~~~~~~~~~~~~~~~~~~~

    Sphinx domain for describing anything.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""

from __future__ import annotations
from typing import Tuple, Dict, Any, Iterator, Type, Set, List, TYPE_CHECKING

from docutils.nodes import Element, literal, Text

from sphinx.addnodes import pending_xref
from sphinx.domains import Domain, ObjType
from sphinx.util import logging
from sphinx.util.nodes import make_refnode
if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import RoleFunction

from .schema import Schema, Object
from .directives import AnyDirective
from .roles import AnyRole
from .indices import AnyIndex

logger = logging.getLogger(__name__)


class AnyDomain(Domain):
    """
    The Any domain for describing anything.
    """

    #: Domain name: should be short, but unique
    name:str = 'any'
    #: Domain label: longer, more descriptive (used in messages)
    label = 'Any'
    #: Type (usually directive) name -> ObjType instance
    object_types:Dict[str,ObjType]= {}
    #: Directive name -> directive class
    directives:Dict[str,Type[AnyDirective]] = {}
    #: Role name -> role callable
    roles:Dict[str,RoleFunction] = {}
    #: A list of Index subclasses
    indices:List[Type[AnyIndex]] = []
    #: AnyDomain specific: Type -> index class
    _indices_for_reftype:Dict[str,Type[AnyIndex]] = {}
    #: AnyDomain specific: Type -> Schema instance
    _schemas:Dict[str,Schema] = {}

    initial_data:Dict[str,Any] = {
        # See property object
        'objects': {},
        # See property references
        'references': {}
    }

    @property
    def objects(self) -> Dict[Tuple[str,str], Tuple[str,str,Object]]:
        """(objtype, objid) -> (docname, anchor, obj)"""
        return self.data.setdefault('objects', {})

    @property
    def references(self) -> Dict[Tuple[str,str,str],Set[str]]:
        """(objtype, objfield, objref) -> set(objid)"""
        return self.data.setdefault('references', {})


    def note_object(self, docname:str, anchor:str, schema:Schema, obj:Object) -> None:
        objtype = obj.objtype
        _, objid = schema.identifier_of(obj)
        objrefs = schema.references_of(obj)
        if (objtype, objid) in self.objects:
            other_docname, other_anchor, other_obj = self.objects[objtype, objid]
            logger.warning(f'duplicate identifier of {obj} at {docname}#{anchor}' +
                           f'other object is {other_obj} at {other_docname}#{other_anchor}')
        logger.debug(f'[any] note object {objtype} {objid} at {docname}#{anchor}, references: {objrefs}')
        self.objects[objtype, objid] = (docname, anchor, obj)
        for objfield, objref in objrefs:
            self.references.setdefault((objtype, objfield, objref), set()).add(objid)


    # Override parent method
    def clear_doc(self, docname:str) -> None:
        objids = set()
        for (objtype, objid), (doc, _, _) in list(self.objects.items()):
            if doc == docname:
                del self.objects[objtype, objid]
                objids.add(objid)
        for (objtype, objfield, objref), ids in list(self.references.items()):
            ids = ids - objids
            if ids:
                self.references[objtype, objfield, objref] = ids
            else:
                del self.references[objtype, objfield, objref]


    # Override parent method
    def resolve_xref(self, env:BuildEnvironment, fromdocname:str,
                     builder:Builder, typ:str, target:str,
                     node:pending_xref, contnode:Element,
                     ) -> Optional[Element]:
        assert isinstance(contnode, literal)

        logger.debug('[any] resolveing xref of %s', (typ, target))
        objtype, objfield = reftype_to_objtype_and_objfield(typ)
        objids = set()
        if objfield:
            # NOTE: To prevent change domain data, dont use ``objids = xxx``
            ids = self.references.get((objtype, objfield, target))
            if ids:
                objids.update(ids)
        else:
            for (t, _, r), ids in self.references.items():
                if t == objtype and r == target:
                    objids.update(ids)

        schema = self._schemas[objtype]
        title = contnode[0].astext()
        has_explicit_title = node['refexplicit']
        newtitle = None

        if not objids:
            # The pending_xref node may be resolved by intersphinx,
            # so not emit warning here, see also warn_missing_reference.
            return None
        elif len(objids) == 1:
            todocname, anchor, obj = self.objects[objtype, objids.pop()]
            if not has_explicit_title:
                newtitle = schema.render_reference(obj)
        else:
            todocname, anchor, = self._get_index_anchor(typ, target)
            if not has_explicit_title:
                newtitle = schema.render_ambiguous_reference(title)
            logger.debug(f'ambiguous {objtype} {target} in {self}, ' +
                        f'ids: {objids} index: {todocname}#{anchor}')

        if newtitle:
            logger.debug(f'[any] rewrite title from {title} to {newtitle}')
            contnode.replace(contnode[0], Text(newtitle))

        return make_refnode(builder, fromdocname, todocname, anchor,
                            contnode, objtype + ' ' + target)


    # Override parent method
    def get_objects(self) -> Iterator[Tuple[str, str, str, str, str, int]]:
        for (objtype, objid), (docname, anchor, _) in self.data['objects'].items():
            yield objid, objid, objtype, docname, anchor, 1


    @classmethod
    def add_schema(cls, schema:Schema) -> None:
        # Add to schemas dict
        cls._schemas[schema.objtype] = schema

        # Generates reftypes(role names) for all referenceable fields
        reftypes = [schema.objtype]
        for name, field, _ in schema.fields_of(None):
            if field.referenceable:
                reftypes.append(objtype_and_objfield_to_reftype(schema.objtype, name))

        # Roles is used for converting role name to corrsponding objtype
        cls.object_types[schema.objtype] = ObjType(schema.objtype, *reftypes)
        cls.directives[schema.objtype] = AnyDirective.derive(schema)
        for r in reftypes:
            # Create role for referencing object (by various fields)
            _, field = reftype_to_objtype_and_objfield(r)
            cls.roles[r] = AnyRole.derive(schema, field)(
                # Emit warning when missing reference (node['refwarn'] = True)
                warn_dangling=True,
                # Inner node (contnode) would be replaced in resolve_xref method,
                # so fix its class.
                innernodeclass=literal,
            )

            index = AnyIndex.derive(schema, field)
            cls.indices.append(index)
            cls._indices_for_reftype[r] = index


    def _get_index_anchor(self, reftype:str, refval:str) -> Tuple[str,str]:
        """
        Return the docname and anchor name of index page. Can be used for ``make_refnode()``.

        .. warning:: This is no public API of sphinx and may broken in future version.
        """
        domain = self.name
        index = self._indices_for_reftype[reftype].name
        return f'{domain}-{index}', f'cap-{refval}'


def warn_missing_reference(app: Sphinx, domain: Domain, node: pending_xref
                           ) -> Optional[bool]:
    if domain and domain.name != AnyDomain.name:
        return None

    objtype, _ = reftype_to_objtype_and_objfield(node['reftype'])
    target = node['reftarget']

    msg = f'undefined {objtype}: {target}'
    logger.warning(msg, location=node, type='ref', subtype=objtype)
    return True


def reftype_to_objtype_and_objfield(reftype:str) -> Tuple[str,Optional[str]]:
    """Helper function for converting reftype(role name) to object infos."""
    reftype = reftype.split('.', maxsplit=1)
    return reftype[0], reftype[1] if len(reftype) == 2 else None


def objtype_and_objfield_to_reftype(objtype:str, objfield:str) -> str:
    """Helper function for converting object infos to reftype(role name)."""
    return objtype + '.' + objfield



"""
    sphinxnotes.any.directives
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Directive implementations.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""
from __future__ import annotations
from typing import Dict, List, Type

from docutils import nodes
from docutils.statemachine import StringList
from docutils.parsers.rst import directives
from docutils.nodes import fully_normalize_name

from sphinx import addnodes
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import make_id, nested_parse_with_titles
from sphinx.util import logging

from .schema import Schema, Object

logger = logging.getLogger(__name__)

class AnyDirective(SphinxDirective):
    """
    Directive to describe anything.  Not used directly,
    but dynamically subclassed to describe specific object.

    The class is modified from sphinx.directives.ObjectDescription
    """

    schema:Schema = None

    # Member of parent
    has_content:bool = True
    required_arguments:int = 0
    optional_arguments:int = 0
    final_argument_whitespace:bool = True
    option_spec:Dict[str,callable] = {}

    @classmethod
    def derive(cls, schema:Schema) -> Type["AnyDirective"]:
        """Generate an AnyDirective child class for describing object."""
        has_content = schema.content is not None

        if not schema.name:
            required_arguments = 0
            optional_arguments = 0
        elif schema.name.required:
            required_arguments = 1
            optional_arguments = 0
        else:
            required_arguments = 0
            optional_arguments = 1

        option_spec = {}
        for name, field in schema.attrs.items():
            if field.required:
                option_spec[name] = directives.unchanged_required
            else:
                option_spec[name] = directives.unchanged

        # Generate directive class
        return type('Any%sDirective' % schema.objtype.title(),
                    (AnyDirective,),
                    {'schema': schema,
                     'has_content': has_content,
                     'required_arguments': required_arguments,
                     'optional_arguments': optional_arguments,
                     'option_spec': option_spec, })


    def _build_object(self) -> Object:
        """Build object information for template rendering."""
        return self.schema.object(name=self.arguments[0] if self.arguments else None,
                                  attrs=self.options,
                                  # Convert docutils.statemachine.ViewList.data -> str
                                  content='\n'.join(list(self.content.data)))


    def _setup_nodes(self, obj:Object,
                   sectnode:nodes.Node,
                   ahrnode:nodes.Node,
                   contnode:nodes.Node) -> None:
        """
        Attach necessary informations to nodes and note them.

        The necessary information contains: domain info, basic attributes for nodes
        (ids, names, classes...), name of anchor, description content and so on.

        :param sectnode: Section node, used as container of the whole object description
        :param ahrnode: Anchor node, used to mark the location of object description
        :param contnode: Content node, which contains the description content
        """
        domainname, objtype = self.name.split(':', 1)
        domain = self.env.get_domain(domainname)

        # Attach domain related info to section node
        sectnode['domain'] = domain.name
        # 'desctype' is a backwards compatible attribute
        sectnode['objtype'] = sectnode['desctype'] = objtype
        sectnode['classes'].append(domain.name)

        # Setup anchor
        _, objid = self.schema.identifier_of(obj)
        ahrid = make_id(self.env, self.state.document, prefix=objtype, term=objid)
        ahrnode['ids'].append(ahrid)
        # Add object name to node's names attribute.
        # 'names' is space-separated list containing normalized reference
        # names of an element.
        name = self.schema.name_of(obj)
        if isinstance(name, str):
            ahrnode['names'].append(fully_normalize_name(name))
        elif isinstance(name, list):
            ahrnode['names'].extend([fully_normalize_name(x) for x in name])
        self.state.document.note_explicit_target(ahrnode)
        # Note object by docu fields
        # FIXME: Cast to AnyDomain
        domain.note_object(self.env.docname, ahrid, self.schema, obj)

        # Parse description
        nested_parse_with_titles(self.state,
                                 StringList(self.schema.render_description(obj)),
                                 contnode)


    def _run_section(self, obj:Object) -> List[nodes.Node]:
        # Get the title of the "section" where the directive is located
        sectnode = self.state.parent
        titlenode = sectnode.next_node(nodes.title)
        if not titlenode or titlenode.parent != sectnode:
            # Title should be direct child of section
            msg = 'Failed to get title of current section'
            logger.warning(msg, location=sectnode)
            sm = nodes.system_message(msg, type='WARNING', level=2, backrefs=[], source='')
            sectnode += sm
            title = ''
        else:
            title = titlenode.astext()
        # Replace the first name "_" with section title
        name = title + obj.name[1:]
        # Object is immutable, so create a new one
        obj = self.schema.object(name=name, attrs=obj.attrs, content=obj.content)
        # NOTE: In _setup_nodes, the anchor node(ahrnode) will be noted by
        # `note_explicit_target` for ahrnode, while `sectnode` is already noted
        # by `note_implicit_target`.
        # Multiple `note_xxx_target` calls to same node causes undefined behavior,
        # so we use `titlenode` as anchor node
        #
        # See https://github.com/sphinx-notes/any/issues/18
        self._setup_nodes(obj, sectnode, titlenode, sectnode)
        # Add all content to existed section, so return nothing
        return []


    def _run_objdesc(self, obj:Object) -> List[nodes.Node]:
        descnode = addnodes.desc()
        # Generate signature node
        title = self.schema.title_of(obj)
        if title is None:
            # Use non-generated object ID as replacement of title
            idfield, objid = self.schema.identifier_of(obj)
            if idfield is not None:
                # ID is not generated
                title = objid
            else:
                # ID is auto-generated, Use '' is better than None
                title = ''
        signode = addnodes.desc_signature(title, '')
        signode += addnodes.desc_name(title, title)
        descnode.append(signode)
        # Generate content node
        contnode = addnodes.desc_content()
        descnode.append(contnode)
        self._setup_nodes(obj, descnode, signode, contnode)
        return [descnode]


    def run(self) -> List[nodes.Node]:
        obj = self._build_object()
        if self.schema.title_of(obj) == '_':
            # If no argument is given, or the first argument is '_',
            # use the section title as object name and anchor,
            # append content nodes to section node
            return self._run_section(obj)
        else:
            # Else, create Sphinx ObjectDescription(sphinx.addnodes.dsec_*)
            return self._run_objdesc(obj)



"""
    sphinxnotes.any.schema
    ~~~~~~~~~~~~~~~~~~~~~~

    Schema and object implementations.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""
from typing import Tuple, Dict, Iterator, List, Set, Optional, Union, Any
from enum import Enum, auto
from dataclasses import dataclass
import pickle
import hashlib

from sphinx.util import logging

from .errors import AnyExtensionError
from .template import Environment as TemplateEnvironment


logger = logging.getLogger(__name__)

class ObjectError(AnyExtensionError):
    pass

class SchemaError(AnyExtensionError):
    pass


@dataclass(frozen=True)
class Object(object):
    objtype:str
    name:str
    attrs:Dict[str,str]
    content:str

    def hexdigest(self) -> str:
        return hashlib.sha1(pickle.dumps(self)).hexdigest()[:7]


@dataclass
class Field(object):
    """
    Describes value constraint of field of Object.

    The value of field can be single or mulitple string.

    :param form: The form of value.
    :param unique: Whether the field is unique.
        If true, the value of field must be unique within the scope of objects
        with same type. And it will be used as base string of object
        identifier.

        Only one unique field is allowed in one object type.

        .. note::

            Duplicated value causes a warning when building documentation,
            and the corresponding object cannot be referenced correctly.
    :param referenceable: Whether the field is referenceable.
        If ture, object can be referenced by field value.
        See :ref:`roles` for more details.

    :param required: Whether the field is required.
        If ture, :py:exc:`ObjectError` will be raised when building documentation
        if the value is no given.
    """

    class Form(Enum):
        "An enumeration represents various string forms."

        #: A single string
        PLAIN = auto()

        #: Mulitple string separated by whitespace
        WORDS = auto()

        #: Mulitple string separated by newline(``\n``)
        LINES = auto()

    form:Form=Form.PLAIN
    unique:bool=False
    referenceable:bool=False
    required:bool=False

    def _as_plain(self, rawval:str) -> str:
        assert self.form == self.Form.PLAIN
        assert rawval is not None
        return rawval


    def _as_words(self, rawval:str) -> List[str]:
        assert self.form == self.Form.WORDS
        assert rawval is not None
        return [x.strip() for x in rawval.split(' ') if x.strip() != '']


    def _as_lines(self, rawval:str) -> List[str]:
        assert self.form == self.Form.LINES
        assert rawval is not None
        return [x.strip() for x in rawval.split('\n') if x.strip() != '']


    def value_of(self, rawval:Optional[str]) -> Union[None,str,List[str]]:
        if rawval is None:
            assert not self.required
            return None

        if self.form == self.Form.PLAIN:
            return self._as_plain(rawval)
        elif self.form == self.Form.WORDS:
            return self._as_words(rawval)
        elif self.form == self.Form.LINES:
            return self._as_lines(rawval)
        else:
            raise NotImplementedError


class Schema(object):
    """
    Schema is used to describe objects, and be able to generate corresponding
    directive, role and index for describing, referencing, and indexing specific
    object.
    """

    #: Template variable name of object type
    TYPE_KEY = 'type'
    #: Template variable name of object name
    NAME_KEY = 'name'
    #: Template variable name of object content
    CONTENT_KEY = 'content'
    #: Template variable name of object title
    TITLE_KEY = 'title'

    # Object type
    objtype:str

    # Object fields
    name:Field
    attrs:Dict[str,Field]
    content:Field

    # Class-wide shared template environment
    # FIXME: can not save jinja template environment because the following error::
    #   Traceback (most recent call last):
    #     File "/usr/lib/python3.9/site-packages/sphinx/cmd/build.py", line 280, in build_main
    #       app.build(args.force_all, filenames)
    #     File "/usr/lib/python3.9/site-packages/sphinx/application.py", line 350, in build
    #       self.builder.build_update()
    #     File "/usr/lib/python3.9/site-packages/sphinx/builders/__init__.py", line 292, in build_update
    #       self.build(to_build,
    #     File "/usr/lib/python3.9/site-packages/sphinx/builders/__init__.py", line 323, in build
    #       pickle.dump(self.env, f, pickle.HIGHEST_PROTOCOL)
    #   _pickle.PicklingError: Can't pickle <function sync_do_first at 0x7f839bc9d790>: it's not the same object as jinja2.filters.sync_do_first
    description_template:str
    reference_template:str
    missing_reference_template:str
    ambiguous_reference_template:str

    def __init__(self, objtype:str,
                 name:Optional[Field]=Field(unique=True, referenceable=True),
                 attrs:Dict[str,Field]={},
                 content:Optional[Field]=Field(),
                 description_template:str='{{ content }}',
                 reference_template:str='{{ title }}',
                 missing_reference_template:str='{{ title }} (missing reference)',
                 ambiguous_reference_template:str='{{ title }} (disambiguation)') -> None:
        """Create a Schema instance.

        :param objtype: The unique type name of object, it will be used as
            basename of corresponding :ref:`directives`, :ref:`roles` and
            :ref:`indices`
        :param name: Constraints of optional object name
        :param attrs: Constraints of object attributes
        :param content: Constraints of object content
        :param description_template: See :ref:`description-template`
        :param reference_template: See :ref:`reference-template`
        :param missing_reference_template: See :ref:`reference-template`
        :param ambiguous_reference_template: See :ref:`reference-template`
        """

        self.objtype = objtype
        self.name = name
        self.attrs = attrs
        self.content = content
        self.description_template = description_template
        self.reference_template = reference_template
        self.missing_reference_template = missing_reference_template
        self.ambiguous_reference_template = ambiguous_reference_template

        # Check attrs constraint
        has_unique = False
        for field in [self.name, self.content] + list(self.attrs.values()):
            if has_unique and field.unique:
                raise SchemaError('only one unique field is allowed in schema')
            else:
                has_unique = field.unique


    def object(self, name:Optional[str], attrs:Dict[str,str], content:Optional[str]) -> Object:
        """Generate a object"""
        obj = Object(objtype=self.objtype,
                     name=name,
                     attrs=attrs,
                     content=content)
        for name, field, val in self.fields_of(obj):
            if field.required and val is None:
                raise ObjectError(f'value of field {name} is none while it is required')
        return obj


    def fields_of(self, obj:Object) -> Iterator[Tuple[str,Field,Union[None,str,List[str]]]]:
        """
        Helper method for returning all fields of object and its raw values.
        -> Iterator[field_name, field_instance, field_value],
        while the field_value is Union[string_value, string_list_value].
        """
        if self.name:
            yield (self.NAME_KEY, self.name, self.name.value_of(obj.name) if obj else None)
        for name, field in self.attrs.items():
            yield (name, field, field.value_of(obj.attrs.get(name)) if obj else None)
        if self.content:
            yield (self.CONTENT_KEY, self.content, self.content.value_of(obj.content) if obj else None)


    def name_of(self, obj:Object) -> Union[None,str,List[str]]:
        assert obj
        return self.content.value_of(obj.name)


    def attrs_of(self, obj:Object) -> Dict[str,Union[None,str,List[str]]]:
        assert obj
        return {k: f.value_of(obj.attrs.get(k)) for k, f in self.attrs.items()}


    def content_of(self, obj:Object) -> Union[None,str,List[str]]:
        assert obj
        return self.content.value_of(obj.content)


    def identifier_of(self, obj:Object) -> Tuple[Optional[str],str]:
        """
        Return unique identifier of object.
        If there is not any unique field, return (None, obj.hexdigest()) instead.
        """
        assert obj
        for name, field, val in self.fields_of(obj):
            if not field.unique:
                continue
            if val is None:
                break
            elif isinstance(val, str):
                return name, val
            elif isinstance(val, list) and len(val) > 0:
                return name, val[0]
            return name, val
        return None, obj.hexdigest()


    def title_of(self, obj:Object) -> Optional[str]:
        """Return title (display name) of object."""
        assert obj
        name = self.name.value_of(obj.name)
        if isinstance(name, str):
            return name
        elif isinstance(name, list) and len(name) > 0:
            return name[0]
        else:
            return None


    def references_of(self, obj:Object) -> Set[Tuple[str,str]]:
        """Return all references (referenceable fields) of object"""
        assert obj
        refs = []
        for name, field, val in self.fields_of(obj):
            if not field.referenceable:
                continue
            if val is None:
                continue
            elif isinstance(val, str):
                refs.append((name, val))
            elif isinstance(val, list):
                refs += [(name, x) for x in val]
        return set(refs)


    def _context_without_object(self) -> Dict[str,Union[str,List[str]]]:
        return  {
            self.TYPE_KEY: self.objtype,
        }


    def _context_of(self, obj:Object) -> Dict[str,Union[str,List[str]]]:
        context = self._context_without_object()

        def set_if_not_none(key:str, val:Union[str,List[str]]) -> None:
            if val is not None:
                context[key] = val
        set_if_not_none(self.NAME_KEY, self.name_of(obj))
        set_if_not_none(self.TITLE_KEY, self.title_of(obj))
        set_if_not_none(self.CONTENT_KEY, self.content_of(obj))
        for key, val in self.attrs_of(obj).items():
            set_if_not_none(key, val)

        return context


    def render_description(self, obj:Object) -> List[str]:
        assert obj
        tmpl = TemplateEnvironment().from_string(self.description_template)
        description = tmpl.render(self._context_of(obj))
        logger.debug('[any] render description template %s: %s' %
                    (self.description_template, description))
        return description.split('\n')


    def render_reference(self, obj:Object) -> str:
        assert obj
        tmpl = TemplateEnvironment().from_string(self.reference_template)
        reference = tmpl.render(self._context_of(obj))
        logger.debug('[any] render references template %s: %s',
                    self.reference_template, reference)
        return reference


    def _render_reference_without_object(self, explicit_title:str,
                                         reference_template:str) -> str:
        context = self._context_without_object()
        context[self.TITLE_KEY] = explicit_title
        tmpl = TemplateEnvironment().from_string(reference_template)
        reference = tmpl.render(context)
        logger.debug('[any] render references template without object %s: %s',
                    reference_template, reference)
        return reference


    def render_missing_reference(self, explicit_title:str) -> str:
        logger.debug('[any] render missing references template %s', explicit_title)
        return self._render_reference_without_object(
            explicit_title, self.missing_reference_template)


    def render_ambiguous_reference(self, explicit_title:str) -> str:
        logger.debug('[any] render ambiguous references template %s', explicit_title)
        return self._render_reference_without_object(
            explicit_title, self.ambiguous_reference_template)


    def __eq__(self, other:Any) -> bool:
        """
        Schema is dynamically created and used in sphinx configuration,
        in order to prevent config changed (lead to sphinx environment rebuild),
        we regard schemas with same attributes are equal.
        """
        if not isinstance(other, Schema):
            return False
        return pickle.dumps(self) == pickle.dumps(other)



"""
    sphinxnotes.any
    ~~~~~~~~~~~~~~~

    Sphinx extension entrypoint.

    :copyright: Copyright 2020 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""

from __future__ import annotations
from typing import List, TYPE_CHECKING

from sphinx.util import logging
if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.config import Config

from .template import Environment as TemplateEnvironment
from .domain import AnyDomain, warn_missing_reference
from .schema import Schema, Field

__title__= 'sphinxnotes-any'
__license__ = 'BSD'
__version__ = '2.3.1'
__author__ = 'Shengyu Zhang'
__url__ = 'https://sphinx-notes.github.io/any'
__description__ = 'Sphinx Domain for describing Anything'
__keywords__ = 'documentation, sphinx, extension'

logger = logging.getLogger(__name__)

# Export
Field = Field
Schema = Schema

def _config_inited(app:Sphinx, config:Config) -> None:
    AnyDomain.name = config.any_domain_name
    AnyDomain.label = config.any_domain_name

    # Add schema before registering domain
    for v in app.config.any_schemas:
        AnyDomain.add_schema(v)

    app.add_domain(AnyDomain)


def setup(app:Sphinx):
    """Sphinx extension entrypoint."""

    # Init template environment
    TemplateEnvironment.setup(app)

    app.add_config_value('any_domain_name', 'any', 'env', types=str)
    app.add_config_value('any_schemas', [], 'env', types=List[Schema])
    app.connect('config-inited', _config_inited)
    app.connect('warn-missing-reference', warn_missing_reference)

    return {'version': __version__}



"""
    sphinxnotes.any.roles
    ~~~~~~~~~~~~~~~~~~~~~

    Roles implementations.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""
from __future__ import annotations
from typing import Tuple, Type, Optional

from sphinx.util import logging
from sphinx.roles import XRefRole

from .schema import Schema

logger = logging.getLogger(__name__)

class AnyRole(XRefRole):
    """
    XRefRole child class for referencing to anything. Not used directly,
    but dynamically subclassed to reference to specific .


    TODO: rst reference template
    """

    schema:Schema

    @classmethod
    def derive(cls, schema:Schema, field:str=None) -> Type["AnyRole"]:
        """Generate an AnyRole child class for referencing object."""
        return type('Any%s%sRole' % (schema.objtype.title(), field.title() if field else ''),
                    (cls,),
                    { 'schema': schema })



"""
    sphinxnotes.any.preset
    ~~~~~~~~~~~~~~~~~~~~~~

    Preset sechemas for sphinxnotes.any.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""

from .domain import Schema, Field, dedent

friend = Schema('friend',
                {'avatar': Field(type=str), 'blog': Field(type=str)},
                description_template=dedent("""
                    .. image:: {{ avatar }}
                       :width: 120px
                       :target: {{ blog }}
                       :alt: {{ names[0] }}
                       :align: left

                    :blog: {{ blog }}

                    {{ content | join('\n') }}"""),
                reference_template='@{{ title }}')



"""
    sphinxnotes.any.indices
    ~~~~~~~~~~~~~~~~~~~~~~~

    Object index implementations

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""
from typing import Iterable, List, Tuple, Dict, Set, Optional, Type
from sphinx.domains import Index, IndexEntry

from .schema import Schema

class AnyIndex(Index):
    """
    Index subclass to provide the object reference index.
    """

    schema:Schema
    # TODO: document
    field:Optional[str] = None

    name:str
    localname:str
    shortname:str

    @classmethod
    def derive(cls, schema:Schema, field:str=None) -> Type["AnyIndex"]:
        """Generate an AnyIndex child class for indexing object."""
        if field:
            typ = f'Any{schema.objtype.title()}{field.title()}Index'
            name = schema.objtype + '.' + field
            localname = f'{schema.objtype.title()} {field.title()} Reference Index'
        else:
            typ = f'Any{schema.objtype.title()}Index'
            name = schema.objtype
            localname = f'{schema.objtype.title()} Reference Index'
        return type(typ, (cls,),
                    { 'schema': schema,
                     'field': field,
                     'name': name,
                     'localname': localname,
                     'shortname': 'references',})


    def generate(self, docnames:Iterable[str] = None
                 ) -> Tuple[List[Tuple[str,List[IndexEntry]]], bool]:
        """Override parent method."""
        content = {}  # type: Dict[str, List[IndexEntry]]
        # List of all references
        objrefs = sorted(self.domain.data['references'].items())

        # Reference value -> object IDs
        objs_with_same_ref:Dict[str,Set[str]] = {}

        for (objtype, objfield, objref), objids in objrefs:
            if objtype != self.schema.objtype:
                continue
            if self.field and objfield != self.field:
                continue
            objs = objs_with_same_ref.setdefault(objref, set())
            objs.update(objids)

        for objref, objids in sorted(objs_with_same_ref.items()):
            # Add a entry for objref
            # 1: Entry with sub-entries.
            entries = content.setdefault(objref, [])
            for objid in sorted(objids):
                docname, anchor, obj = self.domain.data['objects'][self.schema.objtype, objid]
                if docnames and docname not in docnames:
                    continue
                name = self.schema.title_of(obj) or objid
                extra = '' if name == objid else objid
                objcont = self.schema.content_of(obj)
                desc = objcont[:50] + '…' if len(objcont) > 50 else objcont
                # 0: Normal entry
                entries.append(IndexEntry(name, 0, docname, anchor, extra, '', desc))

        # sort by first letter
        sorted_content = sorted(content.items())

        return sorted_content, False



"""
    sphinxnotes.any.errors
    ~~~~~~~~~~~~~~~~~~~~~~

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""
from sphinx.errors import ExtensionError

class AnyExtensionError(ExtensionError):
    pass



# -*- coding: utf-8 -*-
"""
    sphinxnotes
    ~~~~~~~~~~~

    This package is a namespace package that contains all extensions
    distributed in the ``sphinx-notes`` distribution.

    :copyright: Copyright 2020 by the Shengyu Zhang.
"""

__import__('pkg_resources').declare_namespace(__name__)




# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import sys
import os
sys.path.insert(0, os.path.abspath('./sphinxnotes'))
import any as proj

with open('README.rst') as f:
    long_desc = f.read()

setup(
    name=proj.__title__,
    version=proj.__version__,
    url=proj.__url__,
    download_url='http://pypi.python.org/pypi/' + proj.__title__,
    license=proj.__license__,
    author=proj.__author__,
    description=proj.__description__,
    long_description=long_desc,
    long_description_content_type = 'text/x-rst',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Documentation',
        'Topic :: Utilities',
    ],
    keywords=proj.__keywords__,
    platforms='any',
    python_requires='>=3',
    packages=find_packages(),
    include_package_data=True,
    # sphinx.util.compat.Directive class is now deprecated in 1.6
    install_requires=['Sphinx>=1.6', 'Jinja2', 'wand'],
    namespace_packages=['sphinxnotes'],
)



"""
    sphinxnotes.any.template
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Jinja template extensions for sphinxnotes.any.

    :copyright: Copyright 2021 Shengyu Zhang
    :license: BSD, see LICENSE for details.
"""

from __future__ import annotations
from typing import Tuple
import os
from os import path
import posixpath
import tempfile
import shutil

from sphinx.util import ensuredir, relative_uri, logging
from sphinx.application import Sphinx
from sphinx.builders import Builder

import jinja2
from wand.image import Image

ANYDIR = '_any'
logger = logging.getLogger(__name__)

class Environment(jinja2.Environment):
    _builder:Builder
    # Exclusive outdir for template filters
    _outdir:str
    # Exclusive srcdir for template filters
    # Actually it is a softlink link to _outdir.
    _srcdir:str


    @classmethod
    def setup(cls, app:Sphinx):
        """You must call this method before instantiating"""
        app.connect('builder-inited', cls._on_builder_inited)
        app.connect('build-finished', cls._on_build_finished)


    @classmethod
    def _on_builder_inited(cls, app:Sphinx):
        cls._builder = app.builder
        cls._outdir = path.join(app.outdir, ANYDIR)
        ensuredir(cls._outdir)
        cls._srcdir = path.join(app.srcdir, ANYDIR)
        if path.islink(cls._srcdir):
            os.unlink(cls._srcdir)
        os.symlink(cls._outdir, cls._srcdir)

        logger.info(f'[any] exclusive srcdir: {cls._srcdir}')
        logger.info(f'[any] exclusive outdir: {cls._outdir}')

    @classmethod
    def _on_build_finished(cls, app:Sphinx, exception):
        os.unlink(cls._srcdir)


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filters['thumbnail'] = self.thumbnail_filter
        self.filters['install'] = self.install_filter
        # self.filters['watermark'] = self._watermark_filter


    def thumbnail_filter(self, imgfn:str) -> str:
        imgfn = self._ensure_rel(imgfn)
        infn, outfn, relfn = self._get_in_out_rel(imgfn)

        if not self._is_outdated(outfn, infn):
            # No need to make thumbnail
            return relfn

        with Image(filename=infn) as img:
            # Remove any associated profiles
            img.thumbnail()
            # If larger than 640x480, fit within box, preserving aspect ratio
            img.transform(resize='640x480>')
            img.save(filename=outfn)
        return relfn


    def install_filter(self, fn:str) -> str:
        """
        Install file to sphinx outdir, return the relative uri of current docname.
        """

        fn = self._ensure_rel(fn)
        src = path.join(self._builder.srcdir, fn)
        target = path.join(self._builder.outdir, ANYDIR, fn)

        if not self._is_outdated(target, src):
            # No need to install file
            return relfn

        ensuredir(path.dirname(target))
        shutil.copy(src, target)
        return self._relative_uri(ANYDIR, fn)


    def watermark_filter(self, imgfn:str) -> str:
        # TODO
        # infn, outfn, relfn = self._get_in_out_rel(imgfn)
        # with Image(filename=infn) as img:
        #     with Image(width=100, height=100,background='#0008', pseudo='caption:@SilverRainZ') as watermark:
        #         # img.watermark(watermark)
        #         watermark.save(filename=outfn)
        # return relfn
        pass


    def _relative_uri(self, *args):
        """Return a relative URL from current docname to ``*args``."""
        docname = self._builder.env.docname
        base = self._builder.get_target_uri(docname)
        return relative_uri(base, posixpath.join(*args))


    def _get_in_out_rel(self, fn:str) -> Tuple[str,str,str]:
        # The pass-in filenames must be relative
        assert not path.isabs(fn)

        infn = path.join(self._builder.srcdir, fn)
        if infn.startswith(self._srcdir):
            # fn is outputted by other filters
            outfn = infn
        else:
            # fn is specified by user
            outfn = path.join(self._srcdir, fn)
            # Make sure output dir exists
            ensuredir(path.dirname(outfn))
        relfn = self._relative_uri(ANYDIR, fn)
        return (infn, outfn, relfn)


    def _ensure_rel(self, fn: str) -> str:
        """Convert site-wide absoulte path to relative path."""
        return path.relpath(fn, '/') if path.isabs(fn) else fn


    def _is_outdated(self, target:str, src: str) -> bool:
        """
        Return whether the target file is older than src file.
        The given filenames must be absoulte paths.
        """

        assert path.isabs(target)
        assert path.isabs(src)

        # If target file not found, regard as outdated
        if not path.exists(target):
            logger.debug(f'[any] {target} is outdated: not found')
            return True

        # Compare mtime
        try:
            targetmtime = path.getmtime(target)
            srcmtime = path.getmtime(src)
            outdated = srcmtime > targetmtime
            if outdated:
                logger.debug(f'[any] {target} is outdated: {srcmtime} > {targetmtime}')
        except Exception as e:
            outdated = True
            logger.debug(f'[any] {target} is outdated: {e}')
        return outdated
